/*
 GPSDO setup TP5 and communicate
 VK3DIP V1.0 25/08/2020
  Initial Version Set TP and Cut Through
 VK3DIP V1.1 20/10/2020
  Add set Nav Mode to Stationary to optimise TP stability.
 VK3DIP V2.0 25/03/2021 
  Add LCD and status etc.
 VK3DIP V2.1 30/03/2021 
  Add SI5351
 VK3DIP V3.0 31/03/2021 
  Major cleanup, change to neoGPS    
 */
 
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27,16,2);  // set the LCD address to 0x27 for a 16 chars and 2 line display

#include <SoftwareSerial.h>
SoftwareSerial GPS(4, 3);

#include <NMEAGPS.h>
#include <GPSport.h>

// The NeoGPS object
static NMEAGPS gps;
static gps_fix fix;

// add SI5351
#include <si5351.h>
#include <Wire.h>
static Si5351 si5351;


byte gps_set_sucess = 0 ;
 
void setup()
{
  // Do LCD Setup
  lcd.init();
  lcd.clear();         
  lcd.backlight();      // Make sure backlight is on

 // Print startup message to the LCD.
  lcd.setCursor(0,0);   //Set cursor to character 0 on line 0
  lcd.print(F("VK3DIP GPSDO V2"));
  
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Start Setup....."));

  // Do GPS Setup
  // Start GPS Serial at default 9600
  GPS.begin(9600); 
  // Start USB/Serial back to PC also 9600
  Serial.begin(9600);

  Serial.println(F(" "));
  Serial.println(F("Neo-7+ GPSDO "));
  Serial.println(F("VK3DIP V2.0 March 2021"));
  
  Serial.println(F("Initialize TP to 100KHz ..."));
  
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Init TP = 100KHz")); //Using all 16chars so no clear needed
  
  //  This UBX command sets the TimePulse to 1HZ unlocked, 100KHz Locked, both 50% duty cycle
  Serial.println(F("Setting uBlox TP5 mode: "));
  uint8_t setTP5[] = {
    0xB5, 0x62, 0x06, 0x31, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00, 0x32, 0x00, 0x00, 0x00, 0x01, 0x00,
    0x00, 0x00, 0xA0, 0x86, 0x01, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00,
    0x00, 0x00, 0xEF, 0x00, 0x00, 0x00, 0xA1, 0xBA };
  while(!gps_set_sucess)
  {
    sendUBX(setTP5, sizeof(setTP5)/sizeof(uint8_t));
    gps_set_sucess=getUBX_ACK(setTP5);
  }
  gps_set_sucess=0;
 Serial.println(F("TP Set successfully "));
  
  //  This UBX command sets the Navigation Mode to Stationary which is optimal for accuracy of Time Pulse
  Serial.println(F("Setting uBlox Nav mode: "));
  
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Set up Nav mode.")); //Using all 16chars so no clear needed
  
  uint8_t setNav[] = {
    0xB5, 0x62, 0x06, 0x24, 0x24, 0x00, 0xFF, 0xFF, 0x02, 0x03, 0x00, 0x00, 0x00, 0x00, 0x10, 0x27,
    0x00, 0x00, 0x05, 0x00, 0xFA, 0x00, 0xFA, 0x00, 0x64, 0x00, 0x2C, 0x01, 0x00, 0x3C, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x4E, 0x60 };
  while(!gps_set_sucess)
  {
    sendUBX(setNav, sizeof(setNav)/sizeof(uint8_t));
    gps_set_sucess=getUBX_ACK(setNav);
  }
  gps_set_sucess=0;
  
  // get rid of all normal measages except GGA
  Serial.println(F("Turn off all measages save GGA "));
  
  lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
  lcd.print(F("Enable GGA only ")); //Using all 16chars so no clear needed
  
 
  Serial.println(F("Switching off NMEA GLL: "));
   uint8_t setGLL[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x2B                   };
   while(!gps_set_sucess)
   {    
   sendUBX(setGLL, sizeof(setGLL)/sizeof(uint8_t));
   gps_set_sucess=getUBX_ACK(setGLL);
   }
   gps_set_sucess=0;
 
   Serial.println(F("Switching off NMEA GSA: "));
   uint8_t setGSA[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x32                   };
   while(!gps_set_sucess)
   {  
   sendUBX(setGSA, sizeof(setGSA)/sizeof(uint8_t));
   gps_set_sucess=getUBX_ACK(setGSA);
   }
   gps_set_sucess=0;
   Serial.println(F("Switching off NMEA GSV: "));
   uint8_t setGSV[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x39                   };
   while(!gps_set_sucess)
   {
   sendUBX(setGSV, sizeof(setGSV)/sizeof(uint8_t));
   gps_set_sucess=getUBX_ACK(setGSV);
   }
   gps_set_sucess=0;
   Serial.print(F("Switching off NMEA RMC: "));
   uint8_t setRMC[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, 0x40                   };
   while(!gps_set_sucess)
   {
   sendUBX(setRMC, sizeof(setRMC)/sizeof(uint8_t));
   gps_set_sucess=getUBX_ACK(setRMC);
   }
      gps_set_sucess=0;
   Serial.print(F("Switching off NMEA VTG: "));
   uint8_t setVTG[] = { 
   0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0xF0, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x05, 0x47                   };
   while(!gps_set_sucess)
   {
   sendUBX(setVTG, sizeof(setVTG)/sizeof(uint8_t));
   gps_set_sucess=getUBX_ACK(setVTG);
   }
    gps_set_sucess=0;
    
     // initialize the Si5351
     bool i2c_found;
     lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
     lcd.print(F("Setup SI5351....")); //Using all 16chars so no clear needed
     Serial.println(F("Setup SI5351...."));

     i2c_found = si5351.init(SI5351_CRYSTAL_LOAD_0PF, 10000000UL, 0); // set for 10MHz ref input
     if(!i2c_found)
       {
          Serial.println(F("Device not found on I2C bus!"));
       }
 // Assume if we are here that 10MHz ref is available
    MySetSI5351();
  
 Serial.println(F("Done now cutthrough ..."));
 Serial.println(F(" "));

// lcd.setCursor(0,1);   //Move cursor to character 0 on line 1
//  lcd.print(F("Set up complete ")); //Using all 16chars so no clear needed
    lcd.clear(); 
  
}
 
void loop()
{

  
  while (gps.available( gpsPort )) {
    fix = gps.read();
    doSomeWork();
  }

}  

 
// Subs and functions 
 
// Send a byte array of UBX protocol to the GPS
void sendUBX(uint8_t *MSG, uint8_t len) {
 // Serial.println("Send UBX Command:");
  for(int i=0; i<len; i++) {
    GPS.write(MSG[i]);
    if ( i % 10 == 0 && i != 0) Serial.println() ;
    PrnHEX(MSG[i]);
  }
  GPS.println();
}

// Print Hex Chars nicely.
 void PrnHEX(uint8_t MyChr) {
 
    Serial.print("0x");
    Serial.print(MyChr < 16 ? "0" : "");
    Serial.print(MyChr, HEX);
    Serial.print(" ");
    
}

// Send Good Time to LCD at current cursor

void MyUTC() {

    if (fix.dateTime.hours < 10) lcd.print(F("0"));
    lcd.print(fix.dateTime.hours);
    lcd.print(F(":"));
   if (fix.dateTime.minutes < 10) lcd.print(F("0"));
   lcd.print(fix.dateTime.minutes);
   lcd.print(F(":"));
    if (fix.dateTime.seconds < 10) lcd.print(F("0"));
    lcd.print(fix.dateTime.seconds);
  

}

// calculate maidenhead gridsquare from given latitude and longitude
// returns 8 char grid locator in elsewhere declared 9 char array loc passed by *pointer

void calcLocator(char *loc, double latitude, double longitude) {
  // Latitude and longitude are in decimal degrees
  // *loc is a reference to a char array declared global elsewhere
  int tempint; // int variable to hold individual chars
  double remainder; // keep track of remainder
 
  // do longitude calcs
  remainder = longitude + 180.0;
  tempint = (int)(remainder / 20.0);
  remainder = remainder - (double)tempint * 20.0;
  loc[0] = (char)tempint + 'A';
   
  tempint = (int)(remainder / 2.0);
  remainder = remainder - 2.0 * (double)tempint;
  loc[2] = (char)tempint + '0';
   
  tempint = (int)(12.0 * remainder);
  remainder = remainder - (double)tempint / 12;
  loc[4] = (char)tempint + 'a';
  
  tempint = (int)(remainder *120);
  loc[6] = (char)tempint + '0';
  
  // do latitude calcs
  remainder = latitude + 90.0;
  tempint = (int)(remainder / 10.0);
  remainder = remainder - (double)tempint * 10.0;
  loc[1] = (char)tempint + 'A';
  
  tempint = (int)(remainder);
  remainder = remainder - (double)tempint;
  loc[3] = (char)tempint + '0';
  
  tempint = (int)(24.0 * remainder);
  remainder = remainder - (double)tempint / 24;
  loc[5] = (char)tempint + 'a';
  
  tempint = (int)(remainder * 240);
  loc[7] = (char)tempint + '0';
  
  //terminate result
   loc[8] = (char)0;
}

static void doSomeWork()
{
  // Got valid data extract the bits if we have them

  
      if (fix.valid.location)
  {
    //  lcd.setCursor(0,0);   //Set cursor to character 0 on line 0
    //  lcd.print(F("LAT:"));
    //  lcd.print(gps.location.lat(), 6);

    //  lcd.setCursor(0,1);   //Set cursor to character 0 on line 1
    //  lcd.print(F("LON:")); 
    //  lcd.print(gps.location.lng(), 6);
    char myloc[9];
  
    lcd.setCursor(0,1);   //Set cursor to character 0 on line 1
    calcLocator(myloc, fix.latitude(), fix.longitude());
    lcd.print(myloc);
   
    }

  if (fix.valid.altitude)
  {
    lcd.setCursor(9,0);   //Set cursor to character 9 on line 0
    lcd.print(fix.altitude());
    lcd.print(F("m"));
 
  }

  if (fix.valid.satellites)
  {
    lcd.setCursor(9,1);   //Set cursor to character 9 on line 1
    lcd.print(F("  "));  // clean out to allow for less than 10 satelites
    lcd.setCursor(9,1);   //Set cursor to character 9 on line 1
    lcd.print(fix.satellites);
  }

//if (fix.valid.hdop)
 // {
 //    lcd.setCursor(12,1);   //Set cursor to character 12 on line 1
 //   lcd.print(fix.hdop);
 // }


  if (fix.valid.time)
  {
      lcd.setCursor(0,0);   //Set cursor to character 0 on line 0
      //lcd.print(F("UTC: ")); 
      MyUTC(); // LCD print time
  }
 
  
  //trace_all( DEBUG_PORT, gps, fix );

} // doSomeWork

// Set frequencies for the SI5351
// Assume called only after 10MHz ref is available.


void MySetSI5351() {
  // Set CLK0 to output 22.625 MHz for Yaesu FT8x7
  si5351.set_freq(2262500000ULL, SI5351_CLK0);

  // Set CLK1 to output 15.6 MHz
//si5351.set_ms_source(SI5351_CLK1, SI5351_PLLB);
 //si5351.set_freq(1560000000ULL, SI5351_CLK1);

    // Set CLK2 to output 30.2 MHz
//  si5351.set_freq(3020000000ULL, SI5351_CLK2);

 delay(500); // delay a bit to stabalize
 
}
 
// Calculate expected UBX ACK packet and parse UBX response from GPS
boolean getUBX_ACK(uint8_t *MSG) {
  uint8_t b;
  uint8_t ackByteID = 0;
  uint8_t ackPacket[10];
  unsigned long startTime = millis();
  Serial.println();
  Serial.print(F("Reading ACK response: "));
 Serial.println();
  // Construct the expected ACK packet    
  ackPacket[0] = 0xB5;  // header
  ackPacket[1] = 0x62;  // header
  ackPacket[2] = 0x05;  // class
  ackPacket[3] = 0x01;  // id
  ackPacket[4] = 0x02;  // length
  ackPacket[5] = 0x00;
  ackPacket[6] = MSG[2];  // ACK class
  ackPacket[7] = MSG[3];  // ACK id
  ackPacket[8] = 0;   // CK_A
  ackPacket[9] = 0;   // CK_B
 
  // Calculate the checksums
  for (uint8_t i=2; i<8; i++) {
    ackPacket[8] = ackPacket[8] + ackPacket[i];
    ackPacket[9] = ackPacket[9] + ackPacket[8];
  }
 
  while (1) {
 
    // Test for success
    if (ackByteID > 9) {
      // All packets in order!
      Serial.println(F(" (SUCCESS!)"));
      return true;
    }
 
    // Timeout if no valid response in 3 seconds
    if (millis() - startTime > 3000) { 
      Serial.println(F(" (FAILED!) Try Again"));
      return false;
    }
 
    // Make sure data is available to read
    if (GPS.available()) {
      b = GPS.read();
 
      // Check that bytes arrive in sequence as per expected ACK packet
      if (b == ackPacket[ackByteID]) { 
        ackByteID++;
        PrnHEX(b);
      } 
      else {
        ackByteID = 0;  // Reset and look again, invalid order
      }
 
    }
  }
}
